from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Mapping, Optional, Sequence, Tuple

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple, MappingItemsTuple
from metricflow_semantics.helpers.mapping_helpers import mf_items_to_dict
from metricflow_semantics.instances import InstanceSet, InstanceSetTransform, MdoInstance
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.sql.sql_exprs import SqlColumnReference, SqlColumnReferenceExpression
from typing_extensions import override

from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.sql.sql_plan import SqlSelectColumn

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class CreateSelectColumnsResult(Mergeable):
    """Result class for `CreateSelectColumnsForInstances`."""

    # Columns that should be in the `SELECT` clause.
    select_column_set: SelectColumnSet
    spec_to_associated_columns_mapping_items: MappingItemsTuple[InstanceSpec, AnyLengthTuple[SqlSelectColumn]]

    @staticmethod
    def create(  # noqa: D102
        select_column_set: SelectColumnSet,
        spec_to_associated_columns_mapping: Mapping[InstanceSpec, Sequence[SqlSelectColumn]],
    ) -> CreateSelectColumnsResult:
        return CreateSelectColumnsResult(
            select_column_set=select_column_set,
            spec_to_associated_columns_mapping_items=tuple(
                (spec, tuple(columns)) for spec, columns in spec_to_associated_columns_mapping.items()
            ),
        )

    @cached_property
    def spec_to_columns_mapping(self) -> Mapping[InstanceSpec, Sequence[SqlSelectColumn]]:  # noqa: D102
        return mf_items_to_dict(self.spec_to_associated_columns_mapping_items)

    @override
    def merge(self, other: CreateSelectColumnsResult) -> CreateSelectColumnsResult:
        return CreateSelectColumnsResult(
            select_column_set=self.select_column_set.merge(other.select_column_set),
            spec_to_associated_columns_mapping_items=self.spec_to_associated_columns_mapping_items
            + other.spec_to_associated_columns_mapping_items,
        )

    @classmethod
    @override
    def empty_instance(cls) -> CreateSelectColumnsResult:
        return CreateSelectColumnsResult(
            select_column_set=SelectColumnSet.empty_instance(), spec_to_associated_columns_mapping_items=()
        )

    def get_columns(self, spec_output_order: Sequence[InstanceSpec] = ()) -> AnyLengthTuple[SqlSelectColumn]:
        """Returns the generated columns.

        If `spec_output_order` is specified, use the relative ordering described to determine the column order.
        """
        if len(spec_output_order) > 0:
            accounted_specs: set[InstanceSpec] = set()
            unknown_specs: set[InstanceSpec] = set()
            output_columns: list[SqlSelectColumn] = []
            for spec in spec_output_order:
                columns = self.spec_to_columns_mapping.get(spec)
                if columns is None:
                    unknown_specs.add(spec)
                    continue
                output_columns.extend(columns)
                accounted_specs.add(spec)

            if len(accounted_specs) != len(spec_output_order):
                unaccounted_specs = set()
                for spec, columns in self.spec_to_columns_mapping.items():
                    if spec not in accounted_specs:
                        unaccounted_specs.add(spec)
                logger.error(
                    LazyFormat(
                        "Mismatch between `spec_output_order` and created specs. This is a bug and should be"
                        " investigated, but returning the default ordering to reduce user-facing errors.",
                        accounted_specs=accounted_specs,
                        unaccounted_specs=unaccounted_specs,
                        unknown_specs=unknown_specs,
                        spec_output_order=spec_output_order,
                        spec_to_columns_mapping=self.spec_to_columns_mapping,
                    )
                )
                return self.select_column_set.columns_in_default_order

            return tuple(output_columns)
        else:
            return self.select_column_set.columns_in_default_order


class CreateSelectColumnsForInstances(InstanceSetTransform[CreateSelectColumnsResult]):
    """Create select column expressions that will express all instances in the set.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        output_to_input_column_mapping: Optional[OrderedDict[str, str]] = None,
    ) -> None:
        """Initializer.

        Args:
            table_alias: the table alias to select columns from
            column_resolver: resolver to name columns.
            output_to_input_column_mapping: if specified, use these columns in the input for the given output columns.
        """
        self._table_alias = table_alias
        self._column_resolver = column_resolver
        self._output_to_input_column_mapping = output_to_input_column_mapping or OrderedDict()

    def transform(self, instance_set: InstanceSet) -> CreateSelectColumnsResult:  # noqa: D102
        spec_to_associated_columns: dict[InstanceSpec, AnyLengthTuple[SqlSelectColumn]] = {}

        metric_cols = []
        for metric_instance in instance_set.metric_instances:
            columns = self._make_sql_column_expression(metric_instance)
            for column in columns:
                metric_cols.append(column)
                metric_spec = metric_instance.spec
                # The metric spec used in the dataflow plan contains additional attributes that are different from
                # the metric spec provided in the query spec (e.g. includes filters), so the mapping should use the
                # simplified form.
                simplified_metric_spec = MetricSpec(element_name=metric_spec.element_name, alias=metric_spec.alias)
                spec_to_associated_columns[simplified_metric_spec] = columns

        simple_metric_input_cols = []
        for simple_metric_input_instance in instance_set.simple_metric_input_instances:
            columns = self._make_sql_column_expression(simple_metric_input_instance)
            for column in columns:
                simple_metric_input_cols.append(column)
                spec_to_associated_columns[simple_metric_input_instance.spec] = columns

        dimension_cols = []
        for dimension_instance in instance_set.dimension_instances:
            columns = self._make_sql_column_expression(dimension_instance)
            for column in columns:
                dimension_cols.append(column)
                spec_to_associated_columns[dimension_instance.spec] = columns

        time_dimension_cols = []
        for time_dimension_instance in instance_set.time_dimension_instances:
            columns = self._make_sql_column_expression(time_dimension_instance)
            for column in columns:
                time_dimension_cols.append(column)
                spec_to_associated_columns[time_dimension_instance.spec] = columns

        entity_cols = []
        for entity_instance in instance_set.entity_instances:
            columns = self._make_sql_column_expression(entity_instance)
            for column in columns:
                entity_cols.append(column)
                spec_to_associated_columns[entity_instance.spec] = columns

        metadata_cols = []
        for metadata_instance in instance_set.metadata_instances:
            columns = self._make_sql_column_expression(metadata_instance)
            for column in columns:
                metadata_cols.append(column)
                spec_to_associated_columns[metadata_instance.spec] = columns

        group_by_metric_cols = []
        for group_metric_instance in instance_set.group_by_metric_instances:
            columns = self._make_sql_column_expression(group_metric_instance)
            for column in columns:
                group_by_metric_cols.append(column)
                spec_to_associated_columns[group_metric_instance.spec] = columns

        return CreateSelectColumnsResult.create(
            SelectColumnSet.create(
                metric_columns=metric_cols,
                simple_metric_input_columns=simple_metric_input_cols,
                dimension_columns=dimension_cols,
                time_dimension_columns=time_dimension_cols,
                entity_columns=entity_cols,
                group_by_metric_columns=group_by_metric_cols,
                metadata_columns=metadata_cols,
            ),
            spec_to_associated_columns_mapping=spec_to_associated_columns,
        )

    def _make_sql_column_expression(
        self,
        element_instance: MdoInstance,
    ) -> AnyLengthTuple[SqlSelectColumn]:
        """Convert one element instance into a SQL column."""
        # Do a sanity check to make sure that there's a 1:1 mapping between the columns associations generated by the
        # column resolver based on the spec, and the columns that are already associated with the instance.
        expected_column_associations = (self._column_resolver.resolve_spec(element_instance.spec),)
        existing_column_associations = element_instance.associated_columns

        # Dict between the expected column name and the corresponding column in the existing columns
        column_matches: Dict[str, List[str]] = {
            expected_column.column_name: [
                col.column_name
                for col in existing_column_associations
                if col.column_correlation_key == expected_column.column_correlation_key
            ]
            for expected_column in expected_column_associations
        }

        # Assert a 1:1 mapping between expected and existing
        assert all([len(x) == 1 for x in column_matches.values()]), (
            f"Did not find exactly one match for each expected column associations.  "
            f"Expected -> existing mappings: {column_matches}"
        )
        existing_names = set([col.column_name for col in existing_column_associations])
        mapped_names = set()
        mapped_cols: List[str] = []
        for mapped_cols in column_matches.values():
            mapped_names.update([col_name for col_name in mapped_cols])
        assert existing_names == mapped_names, (
            f"Not all existing columns were mapped. Existing: {existing_names}.  Mapped: {mapped_cols}, "
            f"{expected_column_associations} -- {existing_column_associations}"
        )

        select_columns = []
        for expected_name, mapped_cols in column_matches.items():
            input_column_name = mapped_cols[0]
            output_column_name = expected_name

            if output_column_name in self._output_to_input_column_mapping:
                input_column_name = self._output_to_input_column_mapping[output_column_name]
            select_columns.append(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression.create(SqlColumnReference(self._table_alias, input_column_name)),
                    column_alias=output_column_name,
                )
            )
        return tuple(select_columns)


def create_simple_select_columns_for_instance_sets(
    column_resolver: ColumnAssociationResolver,
    table_alias_to_instance_set: OrderedDict[str, InstanceSet],
) -> Tuple[SqlSelectColumn, ...]:
    """Creates select columns for instance sets coming from multiple table as defined in table_alias_to_instance_set.

    Used in cases where you join multiple tables and need to render select columns to access all of those.
    """
    column_set = SelectColumnSet.create()
    for table_alias, instance_set in table_alias_to_instance_set.items():
        column_set = column_set.merge(
            instance_set.transform(
                CreateSelectColumnsForInstances(
                    table_alias=table_alias,
                    column_resolver=column_resolver,
                )
            ).select_column_set
        )

    return column_set.columns_in_default_order



from __future__ import annotations

import functools
import itertools
import logging
from abc import ABC, abstractmethod
from typing import Union

from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.comparison_helpers import ComparisonOtherType, SupportsLessThan

logger = logging.getLogger(__name__)


ComparisonKey = AnyLengthTuple[Union[None, SupportsLessThan]]


@functools.total_ordering
class Comparable(SupportsLessThan, ABC):
    """Mixin class that allows classes of any type to be compared.

    This is useful sorting nodes and edges where the nodes and edges may be different classes.
    Implementing classes should ensure that the `comparison_key` aligns with how `__eq__` is defined.
    """

    @property
    @abstractmethod
    def comparison_key(self) -> ComparisonKey:
        """Return a tuple that can be used for comparing different instances of this class."""
        raise NotImplementedError

    @override
    def __lt__(self, other: ComparisonOtherType) -> bool:
        if not isinstance(other, Comparable):
            return NotImplemented

        self_comparison_key: AnyLengthTuple[ComparisonOtherType] = (
            self.__class__.__module__,
            self.__class__.__qualname__,
            *self.comparison_key,
        )
        other_comparison_key: AnyLengthTuple[ComparisonOtherType] = (
            other.__class__.__module__,
            other.__class__.__qualname__,
            *other.comparison_key,
        )

        for self_key_item, other_key_item in itertools.zip_longest(self_comparison_key, other_comparison_key):
            if self_key_item is None:
                return True
            elif other_key_item is None:
                return False
            elif self_key_item < other_key_item:
                return True
            elif self_key_item > other_key_item:
                return False
            # Must be equal, so compare the next item.

        return False



from __future__ import annotations

import pprint
from collections.abc import Sequence, Set
from dataclasses import fields, is_dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Final, List, Mapping, Optional, Sized, Union

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.helpers.string_helpers import mf_indent
from metricflow_semantics.mf_logging.format_option import PrettyFormatOption
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable


class MetricFlowPrettyFormatter:
    """Creates string representations of objects useful for logging / snapshots."""

    _DEFAULT_FORMAT_OPTION: Final[PrettyFormatOption] = PrettyFormatOption()

    def __init__(
        self,
        format_option: Optional[PrettyFormatOption] = None,
    ) -> None:
        """See mf_pformat() for argument descriptions."""
        self._format_option = (
            format_option if format_option is not None else MetricFlowPrettyFormatter._DEFAULT_FORMAT_OPTION
        )
        max_line_length = self._format_option.max_line_length
        if max_line_length is not None and max_line_length <= 0:
            raise ValueError(f"max_line_length must be > 0 as required by pprint.pformat(). Got {max_line_length}")

    def _handle_sequence_like_obj(
        self, sequence_like_obj: Union[Sequence, Set], remaining_line_length: Optional[int]
    ) -> str:
        """Pretty prints a sequence object i.e. list or tuple.

        Args:
            sequence_like_obj: A sequence-like object, including `Set`.
            remaining_line_length: If specified, try to make the string representation <= this many columns wide.

        Returns:
            A string representation of the sequence. e.g. `(1, 2), [1, 2], {1, 2}`
        """
        include_trailing_comma_for_single_item_in_one_line = False
        if isinstance(sequence_like_obj, tuple):
            left_enclose_str = "("
            right_enclose_str = ")"
            include_trailing_comma_for_single_item_in_one_line = True
        elif isinstance(sequence_like_obj, Sequence):
            left_enclose_str = "["
            right_enclose_str = "]"
        elif isinstance(sequence_like_obj, Set):
            left_enclose_str = "{"
            right_enclose_str = "}"
        else:
            raise RuntimeError(f"Unhandled type: {type(sequence_like_obj)}")

        if len(sequence_like_obj) == 0:
            return f"{left_enclose_str}{right_enclose_str}"

        # See if this object can be printed in one line.
        items_as_str = tuple(
            self._handle_any_obj(sequence_item, remaining_line_length=None) for sequence_item in sequence_like_obj
        )
        line_items = [left_enclose_str]
        if len(items_as_str) > 0:
            line_items.extend([", ".join(items_as_str)])
            if include_trailing_comma_for_single_item_in_one_line and len(items_as_str) == 1:
                line_items.append(",")
        line_items.append(right_enclose_str)
        result_without_width_limit = "".join(line_items)

        if remaining_line_length is None or len(result_without_width_limit) <= remaining_line_length:
            return result_without_width_limit

        # The item can't be printed on one line, so do an indented style like:
        """
        [
            1,
            2,
            ...
        ]
        """

        # Convert each item to a pretty string.
        items_as_str = tuple(
            self._handle_any_obj(
                sequence_item,
                remaining_line_length=max(1, remaining_line_length - len(self._format_option.indent_prefix)),
            )
            for sequence_item in sequence_like_obj
        )
        lines = [left_enclose_str]

        # item_block is similar to
        """
        1,
        2,
        3,
        """

        item_block = ",\n".join(items_as_str) + ","
        # Indent the item_block
        if len(item_block) > 0:
            lines.append(mf_indent(item_block, indent_prefix=self._format_option.indent_prefix))
        lines.append(right_enclose_str)
        return "\n".join(lines)

    def _handle_indented_key_value_item(  # type: ignore[misc]
        self,
        key: Any,
        value: Any,
        key_value_seperator: str,
        is_dataclass_like_object: bool,
        remaining_line_length: Optional[int],
    ) -> str:
        """Convert a key / value for a mapping-like object to a string that should be placed in an indented block.

        Mapping-like objects include dictionaries, dataclasses, and Pydantic models. The output of this method would
        look like:

        "'key': [1, 2, 3]" or "arg=Foo()", etc.

        and the caller of this method would add the actual indent.

        Args:
            key: The object representing the key.
            value: The object representing the value.
            key_value_seperator: The string used to separate the key and the value. e.g. ": " for dicts, "=" for
            dataclasses.
            is_dataclass_like_object: Set this to True if the given value object is a dataclass to handle some printing
            options specific to dataclasses.
            remaining_line_length: If specified, try to make the string representation <= this many columns wide.

        Returns:
            The block that represents the key / value item and goes in between "[" / "]" in the string representation
            of the mapping-like object.
        """
        # See if the string representation can fit on one line. e.g. "'a': [1, 2]"
        result_items_without_limit: List[str] = []
        if is_dataclass_like_object and self._format_option.include_object_field_names:
            result_items_without_limit.append(str(key))
        else:
            result_items_without_limit.append(self._handle_any_obj(key, remaining_line_length=None))
        result_items_without_limit.append(key_value_seperator)
        result_items_without_limit.append(self._handle_any_obj(value, remaining_line_length=None))

        result_without_limit = mf_indent(
            "".join(result_items_without_limit), indent_prefix=self._format_option.indent_prefix
        )
        if remaining_line_length is None or len(result_without_limit) <= remaining_line_length:
            return result_without_limit

        # The string representation can't fit on one line - use multiple lines. e.g.
        """
        'key':
          [1, 2, 3, 4]
        """

        # Create the string for the key.
        result_lines: List[str] = []
        if is_dataclass_like_object and self._format_option.include_object_field_names:
            result_lines.append(str(key) + key_value_seperator)
        else:
            # See if the key can be printed on one line. This depends on the length of the value as the key and the
            # the value as at least the first bit of the value is printed on the same line as the key.
            # e.g.
            # "foo"=[
            #   ...
            # ]
            min_length_of_first_value_line = len(self._handle_any_obj(value, remaining_line_length=1).splitlines()[0])

            key_lines = self._handle_any_obj(
                key,
                remaining_line_length=max(
                    1,
                    remaining_line_length - len(key_value_seperator) - min_length_of_first_value_line,
                ),
            ).splitlines()
            # key_lines would be something like:
            # [
            #     "KeyObject(",
            #     "    a='foo',",
            #     "    b='bar',",
            #     ")",
            # ]

            if len(key_lines) == 1:
                result_lines.append(key_lines[0] + key_value_seperator)
            else:
                # The key needs to be printed in multiple lines. In that case, we want a result where the key value
                # separator is on the last line with the key. e.g.
                #
                # KeyObject(
                #     a='foo',
                #     b='bar',
                # ): ... <value>
                #
                result_lines.extend(key_lines[:-1])
                result_lines.append(key_lines[-1] + key_value_seperator)

        # Combine key and value.

        # Similar to the key, how we print the value depends on whether the value fits on one line or not. e.g.
        # foo=[1, 2, 3]
        #
        # or
        #
        # foo=[
        #   1,
        #   2,
        #   3,
        # ]

        # See if the value fits in the previous line.
        remaining_width_for_value = max(1, remaining_line_length - len(result_lines[-1]))
        value_str = self._handle_any_obj(value, remaining_line_length=remaining_width_for_value)
        value_lines = value_str.splitlines()

        if len(value_lines) <= 1:
            # Value can fit in the previous line
            result_lines[-1] = result_lines[-1] + value_lines[0]
        else:
            # For the multi-line value, we want to print the first line of the value on the same line as the last line
            # of the key.
            result_lines[-1] = result_lines[-1] + value_lines[0]
            result_lines.extend(value_lines[1:])

        return mf_indent("\n".join(result_lines), indent_prefix=self._format_option.indent_prefix)

    def pretty_format_object_by_parts(self, class_name: str, field_mapping: Mapping) -> str:
        """Return the string representation given class name and a mapping of the field names to field values."""
        return self._handle_mapping_like_obj(
            mapping=field_mapping,
            left_enclose_str=class_name + "(",
            key_value_seperator="=",
            right_enclose_str=")",
            is_dataclass_like_object=True,
            remaining_line_length=self._format_option.max_line_length,
        )

    def _handle_mapping_like_obj(
        self,
        mapping: Mapping,
        left_enclose_str: str,
        key_value_seperator: str,
        right_enclose_str: str,
        is_dataclass_like_object: bool,
        remaining_line_length: Optional[int],
    ) -> str:
        """Convert a mapping-like object to a pretty string.

        This class treats dataclasses as mappings where the field / field values are the keys / values.


        Args:
            mapping: The mapping object to convert.
            left_enclose_str: The string used on the left side to enclose the object. e.g. "{" for dicts or "Foo(" for
            dataclasses.
            key_value_seperator: The string used to separate keys and values. e.g. ": " for dicts, or "=" for
            dataclasses.
            right_enclose_str: The string used on the right side to enclose the object. e.g. "}" for dicts or ")" for
            dataclasses.
            is_dataclass_like_object: Flag to indicate whether this is a dataclass as there are some differences in
            formatting those.
            remaining_line_length: If specified, try to make the string representation <= this many columns wide.

        Returns:
            A string representation of the mapping. e.g. "{'a'=[1, 2]}" or "Foo(a=[1, 2])".
        """
        # Skip key / values depending on the pretty-print configuration.
        if is_dataclass_like_object:
            new_mapping = {}
            for key, value in mapping.items():
                if not self._format_option.include_none_object_fields and value is None:
                    continue
                if not self._format_option.include_empty_object_fields and isinstance(value, Sized) and len(value) == 0:
                    continue
                if not self._format_option.include_underscore_prefix_fields and key.startswith("_"):
                    continue
                new_mapping[key] = value
            mapping = new_mapping

        if len(mapping) == 0:
            return f"{left_enclose_str}{right_enclose_str}"

        # Handle case if the string representation fits on one line.
        if remaining_line_length is None or remaining_line_length > 0:
            comma_separated_items: List[str] = []
            for key, value in mapping.items():
                key_value_str_items: List[str] = []

                if is_dataclass_like_object:
                    if self._format_option.include_object_field_names:
                        key_value_str_items.append(str(key))
                        key_value_str_items.append(key_value_seperator)
                else:
                    key_value_str_items.append(self._handle_any_obj(key, remaining_line_length=None))
                    key_value_str_items.append(key_value_seperator)
                key_value_str_items.append(self._handle_any_obj(value, remaining_line_length=None))
                comma_separated_items.append("".join(key_value_str_items))
            result_without_limit = "".join((left_enclose_str, ", ".join(comma_separated_items), right_enclose_str))

            if remaining_line_length is None or len(result_without_limit) <= remaining_line_length:
                return result_without_limit

        # Handle multi-line case.
        mapping_items_as_str = []
        for key, value in mapping.items():
            mapping_items_as_str.append(
                self._handle_indented_key_value_item(
                    key=key,
                    value=value,
                    key_value_seperator=key_value_seperator,
                    is_dataclass_like_object=is_dataclass_like_object,
                    remaining_line_length=(remaining_line_length - len(self._format_option.indent_prefix)),
                )
            )
        lines = [left_enclose_str, ",\n".join(mapping_items_as_str) + ",", right_enclose_str]
        return "\n".join(lines)

    def _handle_any_obj(self, obj: Any, remaining_line_length: Optional[int]) -> str:  # type: ignore
        """Convert any object into a pretty string-representation.

        This is called recursively as sequences and mappings have constituent objects of any type.

        Args:
            obj: The object to convert.
            remaining_line_length: If specified, try to make the string representation <= this many columns wide.

        Returns:
            A pretty string-representation of the object.
        """
        if isinstance(obj, Enum):
            return obj.name

        # Check for strings first as they are also sequences.
        if isinstance(obj, str):
            return self._handle_using_pprint(obj, remaining_line_length=remaining_line_length)

        if isinstance(obj, (Sequence, Set)):
            return self._handle_sequence_like_obj(obj, remaining_line_length=remaining_line_length)

        if isinstance(obj, dict):
            return self._handle_mapping_like_obj(
                obj,
                left_enclose_str="{",
                key_value_seperator=": ",
                right_enclose_str="}",
                is_dataclass_like_object=False,
                remaining_line_length=remaining_line_length,
            )

        if isinstance(obj, MetricFlowPrettyFormattable):
            result = obj.pretty_format(
                PrettyFormatContext(
                    formatter=MetricFlowPrettyFormatter(
                        format_option=self._format_option.with_max_line_length(remaining_line_length)
                    )
                )
            )
            if result is not None:
                return result

        if is_dataclass(obj):
            # dataclasses.asdict() seems to exclude None fields, so doing this instead.
            mapping = {field.name: getattr(obj, field.name) for field in fields(obj)}
            return self._handle_mapping_like_obj(
                mapping,
                left_enclose_str=type(obj).__name__ + "(",
                key_value_seperator="=",
                right_enclose_str=")",
                is_dataclass_like_object=True,
                remaining_line_length=remaining_line_length,
            )

        # For Pydantic-like objects with a `dict`-like method that returns field keys / values.
        # In Pydantic v1, it's `.dict()`, in v2 it's `.model_dump()`.
        # Going with this approach for now to check for a Pydantic model as using `isinstance()` requires more
        # consideration when dealing with Pydantic v1 and Pydantic v2 objects.
        pydantic_dict_method = getattr(obj, "model_dump", None) or getattr(obj, "dict", None)
        if pydantic_dict_method is not None and callable(pydantic_dict_method):
            try:
                # Calling `dict` on a Pydantic model does not recursively convert nested fields into dictionaries,
                # which is what we want. `.model_dump()` does the recursive conversion.
                mapping = dict(obj)
                return self._handle_mapping_like_obj(
                    mapping,
                    left_enclose_str=type(obj).__name__ + "(",
                    key_value_seperator="=",
                    right_enclose_str=")",
                    is_dataclass_like_object=True,
                    remaining_line_length=remaining_line_length,
                )
            except Exception:
                # Fall back to built-in pretty-print in case the dict method can't be called. e.g. requires arguments.
                # Consider logging a warning.
                pass

        if isinstance(obj, Path):
            return str(obj)

        # Any other object that's not handled.
        return self._handle_using_pprint(obj, remaining_line_length=remaining_line_length)

    def _handle_using_pprint(self, obj: Any, remaining_line_length: Optional[int]) -> str:  # type: ignore[misc]
        if remaining_line_length is not None:
            return pprint.pformat(obj, width=remaining_line_length, sort_dicts=False)
        else:
            return pprint.pformat(obj, sort_dicts=False)

    def pretty_format(self, obj: Any) -> str:  # type: ignore[misc]
        """Return a pretty string representation of the object that's suitable for logging."""
        return self._handle_any_obj(obj, remaining_line_length=self._format_option.max_line_length)

    @property
    def format_option(self) -> PrettyFormatOption:
        """Return the formatting option used to create this."""
        return self._format_option


@fast_frozen_dataclass()
class PrettyFormatContext:
    """The context to use for pretty-formatting an object."""

    formatter: MetricFlowPrettyFormatter



from __future__ import annotations

import textwrap

MF_INDENT_2_SPACE = "  "


def mf_newline_join(*args: str) -> str:
    """Convenience function for joining string with newlines."""
    return "\n".join(args)


def mf_indent(message: str, indent_level: int = 1, indent_prefix: str = MF_INDENT_2_SPACE) -> str:  # noqa: D103
    return textwrap.indent(message, prefix=indent_prefix * indent_level)


def mf_dedent(text: str) -> str:
    """Remove leading newlines, dedents, and remove tailing newlines.

    This function simplifies the somewhat-frequently used:

        text = textwrap.dedent(
            [triple quote][backslash]
            Line 0
            Line 1
            [triple quote]
        ).rstrip()

    to:

       text = mf_dedent(
           [triple quote]
           Line 0
           Line 1
           [triple quote]
       )
    """
    return textwrap.dedent(text.lstrip("\n")).rstrip("\n")



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder import (
    MetricFlowPathfinder,
)
from metricflow_semantics.experimental.mf_graph.path_finding.pathfinder_result import GraphTraversalResult
from metricflow_semantics.experimental.ordered_set import OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.recipe_writer_path import (
    AttributeRecipeWriterPath,
)
from metricflow_semantics.experimental.semantic_graph.edges.edge_labels import (
    CumulativeMetricLabel,
    DenyDatePartLabel,
    DenyVisibleAttributesLabel,
)
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    GroupByAttributeLabel,
    KeyAttributeLabel,
    LocalModelLabel,
    MetricLabel,
    MetricTimeLabel,
    SimpleMetricLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraph,
    SemanticGraphEdge,
    SemanticGraphNode,
)
from metricflow_semantics.experimental.semantic_graph.trie_resolver.dunder_name_trie import (
    DunderNameTrie,
)
from metricflow_semantics.helpers.time_helpers import PrettyDuration
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.element_filter import GroupByItemSetFilter

logger = logging.getLogger(__name__)


class DunderNameTrieResolver(ABC):
    """ABC for classes that resolve the available group-by items for a query, returned as a trie."""

    def __init__(  # noqa: D107
        self,
        semantic_graph: SemanticGraph,
        path_finder: MetricFlowPathfinder[SemanticGraphNode, SemanticGraphEdge, AttributeRecipeWriterPath],
    ) -> None:
        self._semantic_graph = semantic_graph
        self._path_finder = path_finder

        # The below are used for convenient autocomplete in the IDE. There are some additional constants that can
        # be shared between the resolvers - consolidation pending.
        self._cumulative_metric_label = CumulativeMetricLabel.get_instance()
        self._deny_date_part_label = DenyDatePartLabel.get_instance()
        self._simple_metric_label = SimpleMetricLabel.get_instance()
        self._local_model_label = LocalModelLabel.get_instance()
        self._metric_time_label = MetricTimeLabel.get_instance()
        self._metric_label = MetricLabel.get_instance()
        self._deny_visible_attributes_label = DenyVisibleAttributesLabel.get_instance()
        self._entity_key_attribute_label = KeyAttributeLabel.get_instance()
        self._virtual_semantic_model_ids = (
            SemanticModelId.get_instance(SemanticModelDerivation.VIRTUAL_SEMANTIC_MODEL_REFERENCE.semantic_model_name),
        )
        self._group_by_attribute_label = GroupByAttributeLabel.get_instance()

    @abstractmethod
    def resolve_trie(
        self, source_nodes: OrderedSet[SemanticGraphNode], element_filter: Optional[GroupByItemSetFilter]
    ) -> TrieResolutionResult:
        """Resolve the trie that represents the available group-by items when querying those nodes together.

        e.g. A query for the `bookings` metric and the `views` metric would have the `Metric(bookings)` node and
        the `Metric(views)` node as source nodes.

        When there are multiple source nodes, the available group-by items are the intersection of the set of items
        available to each source node.
        """
        raise NotImplementedError


@fast_frozen_dataclass()
class TrieCacheKey:
    """A key object to use for caching results."""

    key_nodes: AnyLengthTuple[SemanticGraphNode]
    element_filter: Optional[GroupByItemSetFilter]


@dataclass
class TrieResolutionResult(GraphTraversalResult):
    """A result object that contains some additional fields that describe the performance of resolution.

    Those fields were mostly helpful for debugging - this may be considered for removal.
    """

    duration: PrettyDuration
    dunder_name_trie: DunderNameTrie



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Mapping
from functools import cached_property
from typing import Iterable

from dbt_semantic_interfaces.protocols import Entity
from dbt_semantic_interfaces.type_enums import EntityType
from typing_extensions import override

from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, MutableOrderedSet
from metricflow_semantics.mf_logging.attribute_pretty_format import AttributeMapping, AttributePrettyFormattable

logger = logging.getLogger(__name__)


class EntityLookup(AttributePrettyFormattable):
    """A lookup for entities within a semantic model."""

    def __init__(self, entities: Iterable[Entity]) -> None:  # noqa: D107
        self._entity_name_to_entity: Mapping[str, Entity] = {entity.name: entity for entity in entities}

    @cached_property
    def entity_name_to_type(self) -> Mapping[str, EntityType]:  # noqa: D102
        return {entity_name: entity.type for entity_name, entity in self._entity_name_to_entity.items()}

    @cached_property
    def entity_type_to_names(self) -> Mapping[EntityType, FrozenOrderedSet[str]]:  # noqa: D102
        entity_type_to_names: dict[EntityType, MutableOrderedSet[str]] = defaultdict(MutableOrderedSet)
        for entity_name, entity_type in self.entity_name_to_type.items():
            entity_type_to_names[entity_type].add(entity_name)
        return {entity_type: entity_names.as_frozen() for entity_type, entity_names in entity_type_to_names.items()}

    @cached_property
    @override
    def _attribute_mapping(self) -> AttributeMapping:
        return dict(
            **super()._attribute_mapping,
            **{
                attribute_name: getattr(self, attribute_name)
                for attribute_name in (
                    "entity_name_to_type",
                    "entity_type_to_names",
                )
            },
        )



from __future__ import annotations

from dataclasses import dataclass

from dbt_semantic_interfaces.protocols import MetricInput
from dbt_semantic_interfaces.references import MetricReference

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.metric_lookup import MetricLookup


@dataclass(frozen=True)
class InputMetricDefinitionLocation:
    """Describes the location of input metric of a derived metric."""

    derived_metric_reference: MetricReference
    input_metric_list_index: int

    def get_metric_input(self, metric_lookup: MetricLookup) -> MetricInput:
        """Get the associated MetricInput object that this describes."""
        metric = metric_lookup.get_metric(self.derived_metric_reference)
        metric_inputs = MetricLookup.metric_inputs(metric, include_conversion_metric_input=False)
        if self.input_metric_list_index >= len(metric_inputs):
            raise ValueError(
                LazyFormat(
                    "The metric-input list index is invalid for the given metric",
                    metric=metric,
                    input_metric_list_index=self.input_metric_list_index,
                )
            )

        return metric_inputs[self.input_metric_list_index]



from __future__ import annotations

import logging
from typing import Optional

from typing_extensions import Self, override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.collection_helpers.syntactic_sugar import mf_first_non_none, mf_first_non_none_or_raise
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class PrettyFormatOption(Mergeable):
    """Options for `mf_pformat`.

    max_line_length: If the string representation is going to be longer than this, split into multiple lines.
    indent_prefix: The prefix to use for hierarchical indents.
    include_object_field_names: Include field names when printing objects - e.g. Foo(bar='baz') vs Foo('baz')
    include_none_object_fields: Include fields with a None value - e.g. Foo(bar=None) vs Foo()
    include_empty_object_fields: Include fields that are empty - e.g. Foo(bar=()) vs Foo()
    """

    max_line_length: Optional[int] = 120
    indent_prefix: str = "  "
    include_object_field_names: bool = True
    include_none_object_fields: bool = False
    include_empty_object_fields: bool = False
    include_underscore_prefix_fields: bool = False

    def with_max_line_length(self, max_line_length: Optional[int]) -> PrettyFormatOption:
        """Return a copy of self but with a new value for `max_line_length`."""
        return PrettyFormatOption(
            max_line_length=max_line_length,
            indent_prefix=self.indent_prefix,
            include_object_field_names=self.include_object_field_names,
            include_none_object_fields=self.include_none_object_fields,
            include_empty_object_fields=self.include_empty_object_fields,
        )

    @override
    def merge(self: Self, other: PrettyFormatOption) -> PrettyFormatOption:
        return PrettyFormatOption(
            max_line_length=mf_first_non_none(other.max_line_length, self.max_line_length),
            indent_prefix=mf_first_non_none_or_raise(other.indent_prefix, self.indent_prefix),
            include_object_field_names=mf_first_non_none_or_raise(
                other.include_object_field_names,
                self.include_object_field_names,
            ),
            include_none_object_fields=mf_first_non_none_or_raise(
                other.include_none_object_fields,
                self.include_none_object_fields,
            ),
            include_empty_object_fields=mf_first_non_none_or_raise(
                other.include_empty_object_fields,
                self.include_empty_object_fields,
            ),
            include_underscore_prefix_fields=mf_first_non_none_or_raise(
                other.include_underscore_prefix_fields,
                self.include_underscore_prefix_fields,
            ),
        )

    @classmethod
    @override
    def empty_instance(cls) -> PrettyFormatOption:
        return cls()



from __future__ import annotations

import logging
from typing import Generic, Optional, TypeVar

from metricflow_semantics.collection_helpers.mf_type_aliases import ValueT
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass

logger = logging.getLogger(__name__)


@fast_frozen_dataclass()
class ResultContainer(Generic[ValueT]):
    """Container for a cache result.

    Enables `if result` instead of `if result is not None` to handle falsey result values (e.g. empty set).
    """

    value: ValueT


ResultCacheKeyT = TypeVar("ResultCacheKeyT")


class ResultCache(Generic[ResultCacheKeyT, ValueT]):
    """Cache class to simplify checking / getting / setting the cache for a result.

    Usual pattern:

        result = self._cache.get(cache_key)
        if result is not None:
            return result

        result = compute_result(...)
        self._cache[cache_key] = result
        return result

    Pattern with wrapper:

        result_container = self._cache.get(cache_key)
        return result_container.value if result_container else self._cache.set_and_get(cache_key, compute_result(...))

    There is no lock, so this is intended to be used in cases similar to `cached_property` where the result is equal
    given the cache key. Without the lock, there may be repeated compute, and this shouldn't be used where the caller
    expects the same exact object.

    This is a WIP - there may be easier patterns. This could also be used for instrumenting cache hit rates.
    """

    def __init__(self) -> None:  # noqa: D107
        self._cache_dict: dict[ResultCacheKeyT, ResultContainer[ValueT]] = {}

    def get(self, key: ResultCacheKeyT) -> Optional[ResultContainer[ValueT]]:
        """Returns the cache item for a given key. Also see `ResultContainer`."""
        return self._cache_dict.get(key)

    def set_and_get(self, key: ResultCacheKeyT, value: ValueT) -> ValueT:
        """Set the result for the given key and return the same result.

        This allows the return call to be done in one line.
        """
        self._cache_dict[key] = ResultContainer(value=value)
        return value



from __future__ import annotations

import logging
from abc import ABC
from functools import cached_property
from typing import Optional

from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity
from typing_extensions import override

from metricflow_semantics.collection_helpers.mf_type_aliases import AnyLengthTuple
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.comparable import Comparable, ComparisonKey
from metricflow_semantics.experimental.mf_graph.graph_element import HasDisplayedProperty
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


@fast_frozen_dataclass(order=False)
class AttributeRecipeStep(HasDisplayedProperty, Comparable):
    """A step that should be added to the `AttributeRecipe`.

    A path from a metric / measure node to an attribute node in the semantic graph describes the computation that is
    required to generate the corresponding query. The computation is represented by a `AttributeRecipe`.

    The nodes / edges in the path specify additions to the recipe. For example, an edge from one entity node to
    another entity node would specify a step to join to a semantic model.

    Currently, the recipe stores aggregated properties, and each step generally adds or sets some of those
    properties. However, the recipe can be updated to include more context so that it can be used in the
    `DataflowPlanBuilder`.
    """

    # Adds an element to the dunder name. For example, the edge from the `user` entity node to the
    # `country_latest` attribute node would add the `country_latest` element.
    add_dunder_name_element: Optional[str] = None
    # Adds a property that describes that attribute. For example, the metric-time entity node would add the
    # `METRIC_TIME` property.
    add_properties: Optional[AnyLengthTuple[GroupByItemProperty]] = None
    add_model_join: Optional[SemanticModelId] = None
    add_entity_link: Optional[str] = None

    # Set the element type. For example. visiting a time-dimension entity node would set this to `TIME_DIMENSION`.
    set_element_type: Optional[LinkableElementType] = None
    # To handle the case where time dimensions have a defined grain, the minimum time grain can be set to limit the
    # attribute nodes accessible later in the path.
    set_source_time_grain: Optional[TimeGranularity] = None
    # With `set_min_time_grain`, `set_time_grain_access` can be used to check for valid paths. e.g. the node associated
    # with a time dimension with a `month` defined grain can block access to edges that specify access to `day`.
    set_time_grain_access: Optional[ExpandedTimeGranularity] = None
    set_date_part_access: Optional[DatePart] = None
    # Some edges prevent access of date part attributes (e.g. cumulative metrics)
    set_deny_date_part: Optional[bool] = None

    @override
    @property
    def comparison_key(self) -> ComparisonKey:
        return (
            self.add_dunder_name_element,
            self.add_properties,
            self.add_model_join,
            self.set_source_time_grain.value if self.set_source_time_grain is not None else None,
            self.set_element_type,
            self.add_entity_link,
            self.set_time_grain_access,
            self.set_date_part_access.to_int() if self.set_date_part_access is not None else None,
            self.set_deny_date_part,
        )

    @override
    @cached_property
    def displayed_properties(self) -> AnyLengthTuple[DisplayedProperty]:
        properties = list(super().displayed_properties)

        if self.add_dunder_name_element is not None:
            properties.append(
                DisplayedProperty("add_name", self.add_dunder_name_element),
            )
        for linkable_element_property_addition in self.add_properties or ():
            properties.append(DisplayedProperty("add_prop", linkable_element_property_addition.name))
        if self.add_model_join is not None:
            properties.append(DisplayedProperty("add_model_join", self.add_model_join.model_name))
        if self.set_source_time_grain is not None:
            properties.append(DisplayedProperty("set_min_grain", self.set_source_time_grain.name))
        if self.set_element_type is not None:
            properties.append(DisplayedProperty("add_type", self.set_element_type.name))
        if self.add_entity_link is not None:
            properties.append(DisplayedProperty("add_entity_link", self.add_entity_link))
        if self.set_time_grain_access is not None:
            properties.append(DisplayedProperty("set_time_grain", self.set_time_grain_access.name))
        if self.set_date_part_access is not None:
            properties.append(DisplayedProperty("set_date_part", self.set_date_part_access.name))
        if self.set_deny_date_part is not None:
            properties.append(DisplayedProperty("set_deny_date_part", self.set_deny_date_part))
        return tuple(properties)


class AttributeRecipeStepProvider(HasDisplayedProperty, ABC):
    """Interface for a class that provides a step that can be appended to a recipe."""

    @cached_property
    def recipe_step_to_append(self) -> AttributeRecipeStep:  # noqa: D102
        return AttributeRecipeStep()



from __future__ import annotations

from typing import Sequence

from dbt_semantic_interfaces.implementations.elements.dimension import PydanticDimension, PydanticDimensionTypeParams
from dbt_semantic_interfaces.implementations.elements.entity import PydanticEntity
from dbt_semantic_interfaces.implementations.node_relation import PydanticNodeRelation
from dbt_semantic_interfaces.implementations.semantic_model import PydanticSemanticModel
from dbt_semantic_interfaces.type_enums import DimensionType, EntityType, TimeGranularity

from metricflow_semantics.test_helpers.synthetic_manifest.synthetic_manifest_parameter_set import (
    SyntheticManifestParameterSet,
)


class SimpleMetricSemanticModelGenerator:
    """Helps generate semantic models containing simple-metric inputs.

    Each of the generated semantic models contain an entity common to the semantic models containing dimensions so that any
    measure can be queried by any dimension.
    """

    def __init__(  # noqa: D107
        self,
        parameter_set: SyntheticManifestParameterSet,
    ) -> None:
        self._parameter_set = parameter_set

    def generate_semantic_models(self) -> Sequence[PydanticSemanticModel]:  # noqa: D102
        semantic_models = []
        for semantic_model_index in range(self._parameter_set.simple_metric_semantic_model_count):
            entities = [
                PydanticEntity(
                    name=self._get_primary_entity_name(semantic_model_index),
                    type=EntityType.PRIMARY,
                    description=None,
                    role=None,
                    config=None,
                ),
                PydanticEntity(
                    name=self._parameter_set.common_entity_name,
                    type=EntityType.UNIQUE,
                    description=None,
                    role=None,
                    config=None,
                ),
            ]

            dimensions = [
                PydanticDimension(
                    name="ds",
                    type=DimensionType.TIME,
                    type_params=PydanticDimensionTypeParams(
                        time_granularity=TimeGranularity.DAY,
                    ),
                    description=None,
                    metadata=None,
                    config=None,
                ),
            ]
            semantic_model_name = self.get_semantic_model_name(semantic_model_index)
            semantic_models.append(
                PydanticSemanticModel(
                    name=semantic_model_name,
                    node_relation=PydanticNodeRelation(
                        schema_name="demo",
                        alias=semantic_model_name,
                    ),
                    entities=entities,
                    dimensions=dimensions,
                    defaults=None,
                    description=None,
                    primary_entity=None,
                    metadata=None,
                    config=None,
                )
            )

        return semantic_models

    def get_semantic_model_name(self, semantic_model_index: int) -> str:  # noqa: D102
        return f"simple_metric_model_{semantic_model_index:03}"

    def _get_primary_entity_name(self, semantic_model_index: int) -> str:
        return f"simple_metric_model_{semantic_model_index:03}_primary_entity"



from __future__ import annotations

from dbt_semantic_interfaces.implementations.semantic_manifest import (
    PydanticSemanticManifest,
)
from dbt_semantic_interfaces.transformations.flatten_simple_metrics_with_measure_inputs import (
    FlattenSimpleMetricsWithMeasureInputsRule,
)
from dbt_semantic_interfaces.transformations.names import LowerCaseNamesRule
from dbt_semantic_interfaces.transformations.proxy_measure import CreateProxyMeasureRule
from dbt_semantic_interfaces.transformations.pydantic_rule_set import PydanticSemanticManifestTransformRuleSet
from dbt_semantic_interfaces.transformations.replace_input_measures_with_simple_metrics_transformation import (
    ReplaceInputMeasuresWithSimpleMetricsTransformationRule,
)
from dbt_semantic_interfaces.transformations.semantic_manifest_transformer import (
    PydanticSemanticManifestTransformer,
)


def parse_manifest_from_dbt_generated_manifest(manifest_json_string: str) -> PydanticSemanticManifest:
    """Parse a PydanticSemanticManifest given the generated semantic_manifest json from dbt."""
    raw_model = PydanticSemanticManifest.parse_raw(manifest_json_string)
    # The serialized object in the dbt project does not have all transformations applied to it at
    # this time, which causes failures with input simple-metric input resolution.
    # TODO: remove this transform call once the upstream changes are integrated into our dependency tree
    # TODO: align rules between DSI, here, and MFS (if possible!)
    rule_set = PydanticSemanticManifestTransformRuleSet()
    rules = (
        # Primary
        (LowerCaseNamesRule(),),
        # Secondary
        (
            *rule_set.legacy_measure_update_rules,
            # These individual rules come from rule_set.convert_legacy_measures_to_metrics_rules, but
            # dsi requires AddInputMetricMeasuresRule, and metricflow requires that we do NOT run that rule
            # as it is incompatible with a parser like dbt-core that pre-populates input measures.
            CreateProxyMeasureRule(),
            FlattenSimpleMetricsWithMeasureInputsRule(),
            ReplaceInputMeasuresWithSimpleMetricsTransformationRule(),
            *rule_set.general_metric_update_rules,
        ),
    )
    model = PydanticSemanticManifestTransformer.transform(raw_model, rules)
    return model



from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from typing import Callable, Type, TypeVar

from typing_extensions import dataclass_transform

_T = TypeVar("_T")

logger = logging.getLogger(__name__)


if sys.version_info >= (3, 10):

    @dataclass_transform(kw_only_default=True, order_default=True, frozen_default=True)
    def fast_frozen_dataclass(order: bool = True) -> Callable[[Type[_T]], Type[_T]]:
        """Decorator for creating an immutable dataclass that is faster than `@dataclass(frozen=True)`.

        * Calling `hash()` on a frozen dataclass always computes the hash from the fields.
        * `hash()` performance can be improved by caching the hash value.
        * A cached hash can help improve lookups times in sets / dicts.
        * This uses a regular, mutable dataclass under the hood, with an update to the hash method.
        * Since MF relies on `mypy`, this uses the `dataclass_transform` decorator to statically check that the code
          does not make mutations.
        * This also improve class instantiation times as there is a performance penalty with creating a frozen dataclass.
        * All fields / recursive fields must be immutable.
        * Includes default arguments for creating dataclasses.

        """

        def _make_dataclass_function(inner_cls: Type[_T]) -> Type[_T]:
            # noinspection PyArgumentList
            cls_to_return = dataclass(  # type: ignore[call-overload]
                inner_cls, kw_only=True, order=order, unsafe_hash=True
            )
            cls_to_return.__mf_previous_hash_method = cls_to_return.__hash__

            def _new_hash_method(self) -> int:  # type: ignore[no-untyped-def]
                # Since the object fields are immutable, not using a lock to reduce overhead.
                cached_hash = self.__mf_cached_hash
                if cached_hash is not None:
                    return cached_hash

                cached_hash = self.__mf_previous_hash_method()
                self.__mf_cached_hash = cached_hash
                return cached_hash

            cls_to_return.__mf_cached_hash = None
            cls_to_return.__hash__ = _new_hash_method

            return inner_cls

        return _make_dataclass_function

else:

    @dataclass_transform(order_default=True, frozen_default=True)
    def fast_frozen_dataclass(order: bool = True) -> Callable[[Type[_T]], Type[_T]]:
        """Similar to above but without `kw_args` as it was added in Python 3.10."""

        def _make_dataclass_function(inner_cls: Type[_T]) -> Type[_T]:
            # noinspection PyArgumentList
            cls_to_return = dataclass(inner_cls, order=order, unsafe_hash=True)  # type: ignore[call-overload]
            cls_to_return.__mf_previous_hash_method = cls_to_return.__hash__

            def _new_hash_method(self) -> int:  # type: ignore[no-untyped-def]
                # Since the object fields are immutable, not using a lock to reduce overhead.
                cached_hash = self.__mf_cached_hash
                if cached_hash is not None:
                    return cached_hash

                cached_hash = self.__mf_previous_hash_method()
                self.__mf_cached_hash = cached_hash
                return cached_hash

            cls_to_return.__mf_cached_hash = None
            cls_to_return.__hash__ = _new_hash_method

            return inner_cls

        return _make_dataclass_function



from __future__ import annotations

import functools
from enum import Enum

from metricflow_semantics.experimental.comparison_helpers import ComparisonOtherType


@functools.total_ordering
class OrderedEnum(Enum):
    """An `Enum` that can be sorted by definition order.

    * This is useful dataclasses that have enums as fields as the default `Enum` isn't sortable.
    * Considering using the `OrderedEnum` from `aenum`.
    """

    @classmethod
    @functools.cache
    def _enum_to_index(cls) -> dict[OrderedEnum, int]:
        return {member: i for i, member in enumerate(cls)}

    def __lt__(self, other: ComparisonOtherType) -> bool:  # noqa: D105
        if type(other) is not type(self):
            return NotImplemented
        enum_to_index = self._enum_to_index()

        return enum_to_index[self] < enum_to_index[other]



from __future__ import annotations

from typing import FrozenSet, Iterable, Optional

from typing_extensions import Self, override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty


@fast_frozen_dataclass()
class GroupByItemSetFilter(Mergeable):
    """Describes a way to filter the items in a `BaseGroupByItemSet`."""

    # A `None` value for element names means no filtering on element names.
    element_name_allowlist: Optional[FrozenSet[str]] = None
    any_properties_allowlist: FrozenSet[GroupByItemProperty] = GroupByItemProperty.all_properties()
    any_properties_denylist: FrozenSet[GroupByItemProperty] = frozenset()

    @staticmethod
    def create(  # noqa: D102
        element_name_allowlist: Optional[Iterable[str]] = None,
        any_properties_allowlist: Iterable[GroupByItemProperty] = GroupByItemProperty.all_properties(),
        any_properties_denylist: Iterable[GroupByItemProperty] = frozenset(),
    ) -> GroupByItemSetFilter:
        return GroupByItemSetFilter(
            element_name_allowlist=frozenset(element_name_allowlist) if element_name_allowlist is not None else None,
            any_properties_allowlist=frozenset(any_properties_allowlist),
            any_properties_denylist=frozenset(any_properties_denylist),
        )

    def copy(
        self,
        element_name_allowlist: Optional[FrozenSet[str]] = None,
        any_properties_allowlist: Optional[FrozenSet[GroupByItemProperty]] = None,
        any_properties_denylist: Optional[FrozenSet[GroupByItemProperty]] = None,
    ) -> GroupByItemSetFilter:
        """Create a copy of this with the given non-None fields replaced."""
        return GroupByItemSetFilter(
            element_name_allowlist=element_name_allowlist
            if element_name_allowlist is not None
            else self.element_name_allowlist,
            any_properties_allowlist=any_properties_allowlist
            if any_properties_allowlist is not None
            else self.any_properties_allowlist,
            any_properties_denylist=any_properties_denylist
            if any_properties_denylist is not None
            else self.any_properties_denylist,
        )

    @override
    def merge(self: Self, other: GroupByItemSetFilter) -> GroupByItemSetFilter:
        if self.element_name_allowlist is None and other.element_name_allowlist is None:
            element_names = None
        else:
            element_names = (self.element_name_allowlist or frozenset()).union(
                other.element_name_allowlist or frozenset()
            )
        return GroupByItemSetFilter(
            element_name_allowlist=element_names,
            any_properties_allowlist=self.any_properties_allowlist.union(other.any_properties_allowlist),
            any_properties_denylist=self.any_properties_denylist.union(other.any_properties_denylist),
        )

    @classmethod
    @override
    def empty_instance(cls) -> GroupByItemSetFilter:
        return GroupByItemSetFilter.create()

    def without_element_name_allowlist(self) -> GroupByItemSetFilter:
        """Return this filter without the `element_name_allowlist` filter set."""
        return GroupByItemSetFilter.create(
            element_name_allowlist=None,
            any_properties_denylist=self.any_properties_denylist,
            any_properties_allowlist=self.any_properties_allowlist,
        )

    def allow(self, element_name: Optional[str], element_properties: Optional[Iterable[GroupByItemProperty]]) -> bool:
        """Return true if this allows an item with the given name and properties.

        `None` can be specified in cases of incomplete context.
        """
        if element_name is not None:
            allowed_element_name_set = self.element_name_allowlist
            if allowed_element_name_set is not None and element_name not in allowed_element_name_set:
                return False

        if element_properties is not None:
            if len(self.any_properties_allowlist.intersection(element_properties)) == 0:
                return False
            denied_property_set = self.any_properties_denylist
            if denied_property_set and len(denied_property_set.intersection(element_properties)) > 0:
                return False

        return True



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class SyntheticManifestParameterSet:
    """Describes how to generate a synthetic manifest for performance testing.

    Goals are:
    * Allow modeling of similar patterns seen in production manifests.
    * Make generation straightforward.
    * Minimize the number of parameters required.

    Notes:
    * The synthetic manifest groups semantic models into two types - ones associated with simple metrics,
      and others containing dimensions.
    * A dimension with the same name does not appear in multiple semantic models.
    * Al semantic models contain a common entity so that any simple metric can be queried by any dimension.
    * The metric `depth` describes the number of hops that are required to get to the simple metric when following the
      definition tree.
    * Metrics at `depth=0` are simple metrics. Metrics at other depth values are derived.
    * Each metric is defined using all possible metrics at a lower depth.
    * The number of metrics that are generated with a given `depth` is called the `width`.
    * A random seed can be added later.
    """

    # The number of semantic models to generate that are associated with simple metrics.
    simple_metric_semantic_model_count: int
    # For each semantic model for simple metrics, the count of associated simple metrics.
    simple_metrics_per_semantic_model: int

    # The count of semantic models to generate that contain dimensions.
    dimension_semantic_model_count: int
    # For each semantic model containing dimension, the count of dimensions that it should contain.
    categorical_dimensions_per_semantic_model: int

    # See class docstring.
    max_metric_depth: int
    max_metric_width: int

    # The number of saved queries to generate and the number of elements in each.
    saved_query_count: int
    metrics_per_saved_query: int
    categorical_dimensions_per_saved_query: int

    # The name of the entity that is common to all semantic models.
    common_entity_name: str = "common_entity"



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.sql.sql_cte_node import SqlCteAliasMapping
from metricflow.sql.sql_plan import SqlPlanNode, SqlPlanNodeVisitor, SqlSelectColumn
from metricflow.sql.sql_select_node import SqlSelectStatementNode
from metricflow.sql.sql_table_node import SqlTableNode


@dataclass(frozen=True, eq=False)
class SqlSelectTextNode(SqlPlanNode):
    """An SQL select query that can go in the FROM clause.

    Attributes:
        select_query: The SQL select query to include in the FROM clause.
    """

    select_query: str

    @staticmethod
    def create(select_query: str) -> SqlSelectTextNode:  # noqa: D102
        return SqlSelectTextNode(
            parent_nodes=(),
            select_query=select_query,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_PLAN_QUERY_FROM_CLAUSE_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D102
        return "Read From a Select Query"

    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_query_from_clause_node(self)

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D102
        return None

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @override
    def copy(self) -> SqlSelectTextNode:
        return SqlSelectTextNode(
            parent_nodes=tuple(node.copy() for node in self.parent_nodes), select_query=self.select_query
        )



from __future__ import annotations

import logging
from abc import ABC
from functools import cached_property

from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from typing_extensions import override

from metricflow_semantics.experimental.dataclass_helpers import fast_frozen_dataclass
from metricflow_semantics.experimental.mf_graph.comparable import ComparisonKey
from metricflow_semantics.experimental.mf_graph.formatting.dot_attributes import (
    DotColor,
    DotNodeAttributeSet,
)
from metricflow_semantics.experimental.mf_graph.graph_labeling import MetricFlowGraphLabel
from metricflow_semantics.experimental.mf_graph.node_descriptor import MetricFlowGraphNodeDescriptor
from metricflow_semantics.experimental.ordered_set import FrozenOrderedSet, OrderedSet
from metricflow_semantics.experimental.semantic_graph.attribute_resolution.attribute_recipe_step import (
    AttributeRecipeStep,
)
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.node_labels import (
    ComplexMetricLabel,
    ConfiguredEntityLabel,
    JoinedModelLabel,
    LocalModelLabel,
    MetricLabel,
    MetricTimeLabel,
    SimpleMetricLabel,
    TimeClusterLabel,
    TimeDimensionLabel,
)
from metricflow_semantics.experimental.semantic_graph.sg_constant import ClusterNameFactory
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import SemanticGraphNode
from metricflow_semantics.experimental.semantic_graph.sg_node_grouping import SemanticGraphNodeTypedCollection
from metricflow_semantics.experimental.singleton import Singleton
from metricflow_semantics.model.linkable_element_property import GroupByItemProperty
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType

logger = logging.getLogger(__name__)


@fast_frozen_dataclass(order=False)
class ConfiguredEntityNode(SemanticGraphNode, Singleton):
    """Represents an `entity` element as configured in a semantic model / manifest.

    This node is named "configured" to avoid confusion between entities in the semantic manifest and entities in the
    semantic graph.
    """

    entity_name: str
    model_id: SemanticModelId

    @classmethod
    def get_instance(cls, entity_name: str, model_id: SemanticModelId) -> ConfiguredEntityNode:  # noqa: D102
        return cls._get_instance(entity_name=entity_name, model_id=model_id)

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.entity_name, self.model_id)

    @override
    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        dot_node = super(ConfiguredEntityNode, self).as_dot_node(include_graphical_attributes)
        if include_graphical_attributes:
            dot_node = dot_node.with_attributes(color=DotColor.CORNFLOWER_BLUE)
        return dot_node

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"{self.model_id}.{self.entity_name}",
            cluster_name=ClusterNameFactory.CONFIGURED_ENTITY,
        )

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((ConfiguredEntityLabel.get_instance(),))

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_entity_link=self.entity_name,
            add_dunder_name_element=self.entity_name,
        )

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.configured_entity_nodes.add(self)


@fast_frozen_dataclass(order=False)
class JoinedModelNode(SemanticGraphNode, Singleton):
    """An entity that represents the attributes accessible from a semantic model when the name includes an entity link.

    In the query interface, the description for a group-by item (e.g. with the dunder name, `listing__country_latest`
    `listing` is the entity link).

    An entity link often means a join between semantic models, but not always. For example, dimensions
    can be queried with the associated primary entity name as the entity link even when the query does not require
    joining semantic models.

    To capture this behavior, semantic models are represented with 2 seperate nodes.
    """

    model_id: SemanticModelId

    @classmethod
    def get_instance(cls, model_id: SemanticModelId) -> JoinedModelNode:  # noqa: D102
        # return JoinedModelNode(model_id=model_id)
        return cls._get_instance(model_id=model_id)

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"JoinedModel({self.model_id})",
            cluster_name=ClusterNameFactory.get_name_for_model(self.model_id),
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return self.model_id.comparison_key

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((JoinedModelLabel.get_instance(),))

    @override
    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        dot_node = super(JoinedModelNode, self).as_dot_node(include_graphical_attributes)
        if include_graphical_attributes:
            dot_node = dot_node.with_attributes(edge_node_priority=2)
        return dot_node

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.joined_model_nodes.add(self)


@fast_frozen_dataclass(order=False)
class LocalModelNode(SemanticGraphNode, Singleton):
    """An entity that represents the attributes accessible from a semantic model without entity links.

    Also see `JoinedModelNode`.
    """

    model_id: SemanticModelId

    @classmethod
    def get_instance(cls, model_id: SemanticModelId) -> LocalModelNode:  # noqa: D102
        # return LocalModelNode(model_id=model_id)
        return cls._get_instance(model_id=model_id)

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"LocalModel({self.model_id})",
            cluster_name=ClusterNameFactory.get_name_for_model(self.model_id),
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return self.model_id.comparison_key

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((LocalModelLabel.get_instance(),))

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(add_model_join=self.model_id)

    @override
    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        dot_node = super(LocalModelNode, self).as_dot_node(include_graphical_attributes)
        if include_graphical_attributes:
            dot_node = dot_node.with_attributes(edge_node_priority=1)
        return dot_node

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.local_model_nodes.add(self)


@fast_frozen_dataclass(order=False)
class TimeDimensionNode(SemanticGraphNode, Singleton):
    """An entity representing a time dimension configured in a semantic model.

    In the semantic graph, time dimensions are represented as entities that are related to the time entity. The time
    entity has edges to the different queryable grain-related attributes (e.g. `day`, `month`).
    """

    dimension_name: str

    @classmethod
    def get_instance(cls, dimension_name: str) -> TimeDimensionNode:  # noqa: D102
        return cls._get_instance(dimension_name=dimension_name)

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"TimeDimension({self.dimension_name})",
            cluster_name=ClusterNameFactory.TIME_DIMENSION,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.dimension_name,)

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_dunder_name_element=self.dimension_name,
            set_element_type=LinkableElementType.TIME_DIMENSION,
        )

    @override
    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        dot_node = super(TimeDimensionNode, self).as_dot_node(include_graphical_attributes)
        if include_graphical_attributes:
            dot_node = dot_node.with_attributes(edge_node_priority=1)
        return dot_node

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((TimeDimensionLabel.get_instance(),))

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.time_dimension_nodes.add(self)


@fast_frozen_dataclass(order=False)
class MetricTimeNode(SemanticGraphNode, Singleton):
    """An entity that represents `metric_time`."""

    @classmethod
    def get_instance(cls) -> MetricTimeNode:  # noqa: D102
        return cls._get_instance()

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name="MetricTime",
            cluster_name=ClusterNameFactory.TIME_DIMENSION,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return ()

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((MetricTimeLabel.get_instance(), TimeDimensionLabel.get_instance()))

    @cached_property
    @override
    def recipe_step_to_append(self) -> AttributeRecipeStep:
        return AttributeRecipeStep(
            add_dunder_name_element=METRIC_TIME_ELEMENT_NAME,
            add_properties=(GroupByItemProperty.METRIC_TIME,),
            set_element_type=LinkableElementType.TIME_DIMENSION,
        )

    @override
    def as_dot_node(self, include_graphical_attributes: bool) -> DotNodeAttributeSet:
        dot_node = super(MetricTimeNode, self).as_dot_node(include_graphical_attributes)
        if include_graphical_attributes:
            dot_node = dot_node.with_attributes(edge_node_priority=1)
        return dot_node

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.metric_time_nodes.add(self)


@fast_frozen_dataclass(order=False)
class TimeNode(SemanticGraphNode, Singleton):
    """An entity representing time.

    Other entities related to time (time dimensions, metric time) have an edge to this node.
    """

    @classmethod
    def get_instance(cls) -> TimeNode:  # noqa: D102
        return cls._get_instance()

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name="TimeEntity",
            cluster_name=ClusterNameFactory.TIME,
        )

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return ()

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((TimeClusterLabel.get_instance(),))

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.time_nodes.add(self)


@fast_frozen_dataclass(order=False)
class MetricNode(SemanticGraphNode, ABC):
    """ABC for nodes that represent a metric."""

    metric_name: str

    @cached_property
    @override
    def comparison_key(self) -> ComparisonKey:
        return (self.metric_name,)

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return FrozenOrderedSet((MetricLabel.get_instance(), MetricLabel.get_instance(self.metric_name)))


@fast_frozen_dataclass(order=False)
class SimpleMetricNode(MetricNode, Singleton):
    """Represents a simple metric."""

    @classmethod
    def get_instance(cls, metric_name: str) -> SimpleMetricNode:  # noqa: D102
        return cls._get_instance(metric_name=metric_name)

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"SimpleMetric({self.metric_name})", cluster_name=ClusterNameFactory.METRIC
        )

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return super(SimpleMetricNode, self).labels.union((SimpleMetricLabel.get_instance(),))

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.simple_metric_nodes.add(self)


@fast_frozen_dataclass(order=False)
class ComplexMetricNode(MetricNode, Singleton):
    """Represents metrics that are defined from other metrics.

    With the replacement of measures by simple metrics, metric types like ratio, cumulative, and conversion are like
    derived metrics. To avoid confusion with the `derived` metric type, the `complex` prefix is used.
    """

    @classmethod
    def get_instance(cls, metric_name: str) -> ComplexMetricNode:  # noqa: D102
        return cls._get_instance(metric_name=metric_name)

    @cached_property
    @override
    def node_descriptor(self) -> MetricFlowGraphNodeDescriptor:
        return MetricFlowGraphNodeDescriptor(
            node_name=f"ComplexMetric({self.metric_name})", cluster_name=ClusterNameFactory.METRIC
        )

    @cached_property
    @override
    def labels(self) -> OrderedSet[MetricFlowGraphLabel]:
        return super(ComplexMetricNode, self).labels.union((ComplexMetricLabel.get_instance(),))

    @override
    def add_to_typed_collection(self, typed_collection: SemanticGraphNodeTypedCollection) -> None:
        typed_collection.complex_metric_nodes.add(self)



from __future__ import annotations

import logging

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums import DimensionType
from typing_extensions import override

from metricflow_semantics.experimental.dsi.model_object_lookup import (
    ModelObjectLookup,
)
from metricflow_semantics.experimental.semantic_graph.builder.subgraph_generator import (
    SemanticSubgraphGenerator,
)
from metricflow_semantics.experimental.semantic_graph.edges.sg_edges import EntityAttributeEdge
from metricflow_semantics.experimental.semantic_graph.model_id import SemanticModelId
from metricflow_semantics.experimental.semantic_graph.nodes.attribute_nodes import (
    AttributeNode,
    CategoricalDimensionAttributeNode,
)
from metricflow_semantics.experimental.semantic_graph.nodes.entity_nodes import (
    JoinedModelNode,
)
from metricflow_semantics.experimental.semantic_graph.sg_interfaces import (
    SemanticGraphEdge,
)

logger = logging.getLogger(__name__)


class CategoricalDimensionSubgraphGenerator(SemanticSubgraphGenerator):
    """Generator that adds edges for categorical dimensions.

    This generator add edges from the joined-model nodes to the relevant categorical-dimension nodes.
    """

    @override
    def add_edges_for_manifest(self, edge_list: list[SemanticGraphEdge]) -> None:
        for lookup in self._manifest_object_lookup.model_object_lookups:
            self._add_edges_for_model(lookup, edge_list)

    def _get_nodes_for_categorical_dimensions(self, lookup: ModelObjectLookup) -> list[AttributeNode]:
        attribute_nodes: list[AttributeNode] = []

        for dimension in lookup.semantic_model.dimensions:
            if dimension.type is DimensionType.CATEGORICAL:
                attribute_nodes.append(CategoricalDimensionAttributeNode.get_instance(dimension.name))
            elif dimension.type is DimensionType.TIME:
                pass
            else:
                assert_values_exhausted(dimension.type)

        return attribute_nodes

    def _add_edges_for_model(self, lookup: ModelObjectLookup, edge_list: list[SemanticGraphEdge]) -> None:
        model_id = SemanticModelId.get_instance(model_name=lookup.semantic_model.name)
        semantic_model_node = JoinedModelNode.get_instance(model_id)
        edge_list.extend(
            [
                EntityAttributeEdge.create(
                    tail_node=semantic_model_node,
                    head_node=attribute_node,
                )
                for attribute_node in self._get_nodes_for_categorical_dimensions(lookup)
            ]
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.query.group_by_item.resolution_dag.input_metric_location import InputMetricDefinitionLocation
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeSet,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class SimpleMetricGroupByItemSourceNode(GroupByItemResolutionNode):
    """Outputs group-by-items for a simple metric.

    Attributes:
        metric_reference: Get the group-by items for this simple metric.
        metric_input_location: Optional[InputMetricDefinitionLocation]
    """

    metric_reference: MetricReference
    metric_input_location: Optional[InputMetricDefinitionLocation]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 0

    @staticmethod
    def create(  # noqa: D102
        simple_metric_reference: MetricReference, metric_input_location: Optional[InputMetricDefinitionLocation]
    ) -> SimpleMetricGroupByItemSourceNode:
        return SimpleMetricGroupByItemSourceNode(
            parent_nodes=(),
            metric_reference=simple_metric_reference,
            metric_input_location=metric_input_location,
        )

    @override
    def accept(self, visitor: GroupByItemResolutionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_simple_metric_node(self)

    @property
    @override
    def description(self) -> str:
        return "Output group-by-items available for this simple metric."

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.SIMPLE_METRIC_GROUP_BY_ITEM_RESOLUTION_NODE

    @property
    @override
    def displayed_properties(self) -> Sequence[DisplayedProperty]:
        return tuple(super().displayed_properties) + (
            DisplayedProperty(
                key="metric_reference",
                value=str(self.metric_reference),
            ),
        )

    @property
    @override
    def ui_description(self) -> str:
        return f"SimpleMetric({repr(self.metric_reference.element_name)})"

    @override
    def _self_set(self) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet(simple_metric_nodes=(self,))



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.sql.sql_cte_node import SqlCteAliasMapping
from metricflow.sql.sql_plan import SqlPlanNode, SqlPlanNodeVisitor, SqlSelectColumn
from metricflow.sql.sql_select_node import SqlSelectStatementNode
from metricflow.sql.sql_table_node import SqlTableNode


@dataclass(frozen=True, eq=False)
class SqlCreateTableAsNode(SqlPlanNode):
    """An SQL node representing a CREATE TABLE AS statement.

    Attributes:
        sql_table: The SQL table to create.
    """

    sql_table: SqlTable

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(sql_table: SqlTable, parent_node: SqlPlanNode) -> SqlCreateTableAsNode:  # noqa: D102
        return SqlCreateTableAsNode(
            parent_nodes=(parent_node,),
            sql_table=sql_table,
        )

    @override
    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_create_table_as_node(self)

    @property
    @override
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @property
    @override
    def description(self) -> str:
        return f"Create table {repr(self.sql_table.sql)}"

    @property
    def parent_node(self) -> SqlPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.SQL_PLAN_CREATE_TABLE_AS_ID_PREFIX

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return self.parent_node.nearest_select_columns(cte_source_mapping)

    @override
    def copy(self) -> SqlCreateTableAsNode:
        return SqlCreateTableAsNode.create(parent_node=self.parent_node.copy(), sql_table=self.sql_table)
